import re
import pathlib
import tempfile
import urllib.parse

import pycountry
import requests
from lxml.etree import fromstring, HTMLParser
from pycldf.sources import Source

INDEX_URL = "https://www.lddjournal.org/articles/"
URLS = {
    'contexts':
        '{}?date_published__date__gte=&date_published__date__lte=&section__pk=55'.format(INDEX_URL),
    'snapshots':
        '{}?date_published__date__gte=&date_published__date__lte=&section__pk=54'.format(INDEX_URL),
}


def get(url, cache, series=None):  # pragma: no cover
    parsed = urllib.parse.urlparse(url)
    if parsed.path.endswith('articles/'):
        assert series
        fname = 'index_{}.html'.format(series)
    else:
        m = re.fullmatch(r'/article/(pub)?id/(?P<aid>[0-9]+)/', parsed.path)
        if m:
            fname = '{}.html'.format(m.group('aid'))
        else:
            m = re.fullmatch(r'/article/id/(?P<aid>[0-9]+)/download/xml/', parsed.path)
            if m:
                fname = '{}.xml'.format(m.group('aid'))
            else:
                raise ValueError(url)
    p = cache / fname
    if fname.startswith('index') or not p.exists():
        p.write_text(requests.get(url).text, encoding='utf8')
    if p.suffix == '.html':
        return fromstring(p.read_text(encoding='utf8'), parser=HTMLParser())
    return fromstring(p.read_bytes())


def iter_items(repos):  # pragma: no cover
    cache = repos.repos / 'build' / 'elpub'
    if not cache.exists():
        cache.mkdir()

    for series, url in URLS.items():
        for div in get(url, cache, series).xpath('.//div'):
            if 'class' in div.attrib and div.attrib['class'] == 'box article':
                xml = None
                metas = get(div.xpath('a')[0].attrib['href'], cache).xpath('head/meta')
                for meta in metas:
                    if meta.attrib.get('name') == 'citation_xml_url':
                        xml = get(meta.attrib['content'], cache)

                        break
                yield scrape_article(
                    series,
                    [(m.attrib['name'], m.attrib['content']) for m in metas if 'name' in m.attrib],
                    xml)


def download(bibfile, log, repos):  # pragma: no cover
    old, new = [], []
    for e in bibfile.iterentries():
        old.append(e.key)
    for item in iter_items(repos):
        if item.id not in old:
            new.append(item)

    if new:
        fname = pathlib.Path(tempfile.gettempdir()) / 'elpub.bib'
        fname.write_text(
            '\n'.join([r.bibtex() for r in sorted(
                new, key=lambda r: int(r.id.replace('ldd', '')))]), encoding='utf8')

        bibfile.update(fname, log=log, keep_old=True)
        bibfile.check(log)
        fname.unlink()


def scrape_article(series, metas, xml):  # pragma: no cover
    meta2bibtex = {
        'citation_journal_title': ('journal', False, None),
        'citation_issn': ('issn', False, None),
        'citation_author': ('author', True, None),
        'citation_title': ('title', False, None),
        'citation_publication_date': ('year', False, lambda s: s.split('-')[0]),
        'citation_volume': ('volume', False, None),
        'citation_issue': ('number', False, None),
        'citation_doi': ('doi', False, None),
    }
    md = {'hhtype': 'overview'}
    for name, content in metas:
        if name in meta2bibtex:
            bname, mult, conv = meta2bibtex[name]
            content = conv(content) if conv else content
            if not mult:
                md[bname] = content
            else:
                if bname in md:
                    md[bname].append(content)
                else:
                    md[bname] = [content]
    md['author'] = ' and '.join(md['author'])
    md['url'] = 'https://doi.org/{}'.format(md['doi'])
    if xml is not None:
        md.update(parse_xml(xml))
    return Source('article', md['doi'].split('/')[-1].replace('.', ''), **md)


def parse_xml(article):  # pragma: no cover
    from lxml.etree import tostring
    inlg = pycountry.languages.get(
        alpha_2=article.attrib['{http://www.w3.org/XML/1998/namespace}lang'])
    abstract = article.xpath('.//{}abstract'.format('trans-' if inlg.alpha_2 != 'en' else ''))[0]
    if '{http://www.w3.org/XML/1998/namespace}lang' in abstract.attrib:
        assert abstract.attrib['{http://www.w3.org/XML/1998/namespace}lang'] == 'en'
    try:
        res = {
        'inlg': '{} [{}]'.format(inlg.name, inlg.alpha_3),
        'abstract': '\n'.join(re.sub(r'\s+', ' ', p.text) for p in abstract.xpath('p') if p.text),
        'subject': '; '.join(kw.text for kw in article.xpath('.//kwd-group/kwd') if kw.text),
        }
    except:
        print(tostring(article).decode('utf8'))
        raise
    props = []
    for e in article.xpath('body/*'):
        if e.tag == 'sec':
            break
        props.append(e)
    for e in props:  # table-wrap or some p
        m = re.search(r'(?P<gc>[a-z]{4}[0-9]{4})', ''.join(e.itertext()))
        if m:
            res['lgcode'] = '[{}]'.format(m.group('gc'))
            break
    return res



import shutil

from pyglottolog.homelands import *


def test_compute(api_copy):
    from pyglottolog import homelands
    if homelands.geo:  # pragma: no cover
        assert compute(api_copy, md)
        assert compute(api_copy, recursive_centroids)

        abcd1233 = api_copy.tree / 'abcd1233'
        abcd1233.mkdir()
        abcd1233.joinpath('md.ini').write_text(
            '# -*- coding: utf-8 -*-\n[core]\nname = family0\nlevel = family')
        shutil.move(str(api_copy.tree / 'abcd1234'), str(abcd1233))

        assert compute(api_copy, md)
        assert compute(api_copy, recursive_centroids)



"""
Computing geo-coordinates for homelands of language groups, i.e. languoids of level `family`.

Various ways of computing "homelands" for language groups have been proposed in the literature
since Sapir 1916. This module provides implementations of some of the simpler algorithms.
"""
import gzip
import json
import random
import typing
import decimal
import pathlib
import collections

try:
    from shapely.geometry import shape, GeometryCollection, MultiPoint, Point
    from shapely.ops import nearest_points
    import pyproj
    geo = True
except ImportError:  # pragma: no cover
    geo = False

import pyglottolog
from pyglottolog.languoids import Languoid

__all__ = ['compute', 'md', 'recursive_centroids']
random.seed(12345)


def _worlds_land_masses_dict():
    res = {}
    for p in pathlib.Path(pyglottolog.__file__).parent.joinpath('data').glob('*.geojson.gz'):
        with gzip.open(p, mode='rt', encoding='utf8') as fp:
            res[p.stem] = GeometryCollection([
                shape(f["geometry"]).buffer(0) for f in json.loads(fp.read())['features']])
    return res


def compute(api: pyglottolog.Glottolog,
            method: typing.Callable[
                [typing.List[Languoid]],
                typing.Dict[str, typing.Tuple[decimal.Decimal, decimal.Decimal]]])\
        -> typing.Dict[str, typing.Tuple[decimal.Decimal, decimal.Decimal]]:
    """
    Compute homelands for applicable Glottolog subgroups using a method implemented in this module
    or any callable with appropriate signature.
    """
    if not geo:  # pragma: no cover
        raise ValueError('Computing homelands requires the "geo" extra, installed via '
                         '"pip install pyglottolog[geo]"')
    return method(_l1_languages_with_coordinates(api))


def geodist(geod, p1, p2):
    return geod.inv(p1[1], p1[0], p2[1], p2[0])[2]


def md(langs: typing.List[Languoid])\
        -> typing.Dict[str, typing.Tuple[decimal.Decimal, decimal.Decimal]]:
    """
    Compute homeland coordinates for a language group (and its subgroups) as described as
    "md" method in "Testing methods of linguistic homeland detection using synthetic data"
    by Søren Wichmann and Taraka Rama
    https://doi.org/10.1098/rstb.2020.0202

    Wichmann and Rama 2021:

        In the third approach, abbreviated ‘md’ for ‘minimal distance’, we compute the average
        distance (as the crow flies) from each language to all the other languages. The location
        of the language that has the smallest average distance to the others is equated with the
        homeland.

    We use the `pyproj.Geod.inv` method to compute the great-circle distance between two points.

    .. seealso: https://pyproj4.github.io/pyproj/stable/api/geod.html
    """
    # Compute minimal distances per group:
    geod = pyproj.Geod(ellps='WGS84')

    grouped_languages = collections.defaultdict(list)
    for lang in langs:
        for _, gc, _ in lang.lineage:
            grouped_languages[gc].append(lang)

    homelands = {}
    for group, lgs in grouped_languages.items():
        if len(lgs) == 1:  # pragma: no cover
            homelands[group] = (lgs[0].latitude, lgs[0].longitude)
            continue

        # We shuffle the coordinates to avoid returning the first minimal-distance location in the
        # given order.
        coords = [(lg.latitude, lg.longitude) for lg in lgs]
        random.shuffle(coords)
        mindist, mincoord = None, None
        for i, coord in enumerate(coords):
            dist = sum(geodist(geod, coord, p) for j, p in enumerate(coords) if i != j)
            if (mindist is None) or (dist < mindist):
                mindist, mincoord = dist, coord
        homelands[group] = mincoord
    return homelands


def recursive_centroids(langs: typing.List[Languoid])\
        -> typing.Dict[str, typing.Tuple[decimal.Decimal, decimal.Decimal]]:
    """
    Recursively compute homelands of subgroups from the homelands of their immediate children in
    the classification.

    1. The homeland of a single language is its geographic coordinate.
    2. The homeland of a set of coordinates (for homelands or languages) is computed as
       nearest point on land of the centroid of the convex hull for the set of coordinates.
    """
    # We compute centroids with shapely in a projection-agnostic way. Thus, we have to make sure
    # to deal with longitudes wrapping around at 180° - which only happens for subgroups of
    # Austronesian.
    def pos_lon(lon, tlgc):
        return lon + 360 if lon < 0 and tlgc == 'aust1307' else lon

    def norm_lon(lon, tlgc):
        return lon - 360 if lon > 180 and tlgc == 'aust1307' else lon

    subgroups = collections.defaultdict(list)
    pref_continents = None
    for lang in langs:
        if not pref_continents:
            pref_continents = {
                'South America': ['southamerica', 'northaerica'],
                'North America': ['northamerica', 'southamerica'],
                'Eurasia': ['asia', 'europe'],
                'Africa': ['africa'],
                'Papunesia': ['oceania'],
                'Australia': ['oceania'],
            }[lang.macroareas[0].name]
        tlgc = lang.lineage[0][1]
        prev = None
        for i, (_, gc, _) in enumerate(reversed(lang.lineage)):
            if i == 0:  # For the immediate parent, we append the coordinate.
                subgroups[(tlgc, gc)].append((lang.latitude, lang.longitude))
            else:  # Otherwise we append the Glottocode of the immediate child.
                subgroups[(tlgc, gc)].append(prev)
            prev = gc

    geod = pyproj.Geod(ellps='WGS84')
    continents = _worlds_land_masses_dict()
    homelands = {}
    while subgroups:
        for (tlgc, gc), coords in list(subgroups.items()):
            # coords is a list of immediate children of the group specified by `gc`, either
            # given as coordinates of languages or homelands or as glottocodes.
            coords = [homelands.get(v, v) for v in coords]
            if any(isinstance(v, str) for v in coords):
                # There are still unresolved homelands for the immediate children. Defer
                # computation until all homelands of children are resolved.
                continue  # pragma: no cover
            # Compute the homeland from the homelands of the children.
            homeland = MultiPoint(
                [(pos_lon(p[1], tlgc), p[0]) for p in coords]).convex_hull.centroid
            homeland = Point(norm_lon(homeland.x, tlgc), homeland.y)
            for _, l in sorted(
                    continents.items(), key=lambda i: i[0] in pref_continents, reverse=True):
                if l.contains(homeland):
                    break  # pragma: no cover
            else:
                nps = [nearest_points(c, homeland)[0] for c in continents.values()]
                nps = [(p, geodist(geod, (p.y, p.x), (homeland.y, homeland.x))) for p in nps]
                homeland = sorted(nps, key=lambda n: n[1])[0][0]
            homelands[gc] = (homeland.y, homeland.x)
            del subgroups[(tlgc, gc)]
    return homelands


def _l1_languages_with_coordinates(api):
    invalid_macroareas = {
        'atla1278': {api.macroareas.northamerica.name},
        'aust1307': {api.macroareas.southamerica.name},
        'indo1319': {
            api.macroareas.northamerica.name,
            api.macroareas.southamerica.name,
            api.macroareas.africa.name,
            api.macroareas.australia.name,
            api.macroareas.pacific.name,
        },
    }
    return [
        lg for lg in api.languoids()
        if lg.latitude is not None
        and lg.lineage  # noqa: W503
        and lg.level == api.languoid_levels.language  # noqa: W503
        and lg.category == api.language_types.spoken_l1.category  # noqa: W503
        and (  # noqa: W503
            (lg.lineage[0][1] not in invalid_macroareas) or  # noqa: W504
            (not invalid_macroareas[lg.lineage[0][1]].intersection(ma.name for ma in lg.macroareas))
        )
    ]


if __name__ == '__main__':  # pragma: no cover
    from pyglottolog import Glottolog
    import sys

    gl = Glottolog(sys.argv[1], cache=True)
    res = compute(gl, md)
    print(len(res))



import shlex
import logging

import pytest

from pyglottolog.__main__ import main


@pytest.fixture
def _main(api_copy, mocker):
    def f(*args, **kw):
        kw.setdefault('log', mocker.Mock())
        kw['test'] = True
        if len(args) == 1 and isinstance(args[0], str):
            # Windows: shlex.split() drops backslashes, e.g. in paths
            #          shlex.split(posix=False) does not parse quotes
            #          However: supprocess.Popen does not use it
            #          Workaround: use pathlib.Path.as_posix()
            # https://stackoverflow.com/questions/33560364/python-windows-parsing-command-lines-with-shlex
            args = shlex.split(args[0])
        main(args=['--repos', str(api_copy.repos)] + list(args), **kw)
    return f


def test_help(_main, capsys):
    _main()
    out, _ = capsys.readouterr()
    assert 'usage' in out


def test_no_repos(tmpdir):
    with pytest.raises(SystemExit):
        main(args=['--repos', str(tmpdir.join('x')), 'show', 'abcd1235'])


def test_show(capsys, _main):
    _main('show **a:key**')
    assert '@misc' in capsys.readouterr()[0]

    _main('show a:key')
    assert '@misc' in capsys.readouterr()[0]

    _main('show abcd1236')
    assert 'Classificat' in capsys.readouterr()[0]


def test_edit(mocker, _main):
    mocker.patch('pyglottolog.commands.edit.subprocess')
    _main('edit abcd1236')


def test_create(capsys, _main, api_copy):
    with pytest.raises(SystemExit):
        _main('create abcd1249 "new name" language')

    _main('create abcd1234 "new name" language')
    assert 'Info written' in capsys.readouterr()[0]
    assert 'new name' in [c.name for c in api_copy.languoid('abcd1234').children]

    with pytest.raises(SystemExit):
        _main('create {0} "new name" language'.format(
            (api_copy.repos / 'languoids' / 'tree' / 'abcd1249').as_posix()))

    _main('create {0} "new name" language'.format(
        (api_copy.repos / 'languoids' / 'tree' / 'abcd1234').as_posix()))


def test_fts(capsys, _main):
    with pytest.raises(SystemExit):
        _main('refsearch "Harzani year:1334"')

    _main('searchindex')
    _main('refsearch "Harzani year:1334"')
    assert "'Abd-al-'Ali Karang" in capsys.readouterr()[0]

    _main('langsearch id:abcd*')
    assert "abcd1234" in capsys.readouterr()[0]

    _main('langsearch classification')
    assert "abcd1234" in capsys.readouterr()[0]


def test_metadata(capsys, _main):
    _main('langdatastats')
    assert "longitude" in capsys.readouterr()[0]


def test_tree(capsys, _main):
    with pytest.raises(SystemExit):
        _main('tree')

    with pytest.raises(SystemExit):
        _main('tree xyz')

    _main('tree abc --maxlevel language')
    out, _ = capsys.readouterr()
    assert 'language' in out
    assert 'dialect' not in out

    _main('tree abcd1234 --newick --maxlevel 1')
    _main('tree abcd1235 --newick --maxlevel 5')
    assert 'language' in capsys.readouterr()[0]
    _main('tree abcd1234 --newick --maxlevel language')
    out, _ = capsys.readouterr()
    assert out.splitlines()[-1] == \
           "('language [abcd1235][abc]-l-':1," \
           "'language2 [abcd1237]-l-':1)'family [abcd1234][aaa]':1;"


def test_languoids(capsys, _main, tmp_path):
    _main('languoids --output={0}'.format(tmp_path.as_posix()))
    out, _ = capsys.readouterr()
    assert '-metadata.json' in out
    assert tmp_path.glob('*.csv')


def test_htmlmap(_main, tmp_path):
    _main('htmlmap --output {0} --min-langs-for-legend 1'.format(tmp_path.as_posix()))
    assert tmp_path.joinpath('glottolog_map.html').exists()
    tmp_path.joinpath('glottocodes').write_text('abcd1234\nabcd1235,5,8\n', encoding='utf8')
    _main('htmlmap --output {0} --glottocodes {1}'.format(tmp_path.as_posix(),
                                                          (tmp_path / 'glottocodes').as_posix()))

    with pytest.raises(SystemExit):
        _main('htmlmap --output {0}'.format((tmp_path /'xyz').as_posix()))

    with pytest.raises(SystemExit):
        _main('htmlmap --glottocodes {0}'.format(tmp_path.joinpath('xyz')))


def test_iso2codes(_main, tmp_path):
    _main('iso2codes --output {0}'.format(tmp_path.as_posix()))
    assert tmp_path.joinpath('iso2glottocodes.csv').exists()




from pyglottolog.config import Config, AES, DocumentType, Macroarea


def test_aes(api_copy):
    aes = Config.from_ini(api_copy.repos / 'config' / 'aes_status.ini', AES)
    assert aes.safe < aes.definite
    assert aes.vulnerable == aes.get('threatened')


def test_macroarea(api_copy):
    ma = Config.from_ini(api_copy.repos / 'config' / 'macroareas.ini', Macroarea)
    assert ma.__defaults__['description']
    assert ma.eurasia.geojson is None


def test_doctype(api_copy):
    dt = Config.from_ini(api_copy.repos / 'config' / 'document_types.ini', DocumentType)
    assert dt.grammar > dt.grammar_sketch


def test_editors(api_copy):
    assert len([e for e in api_copy.editors.values() if not e.current]) == 2
    assert len([e for e in api_copy.editors.values() if e.current]) == 3



import pytest

from pyglottolog.links import wikidata


def test_wikidata(mocker, api_copy):
    mocker.patch('pyglottolog.links.wikidata.confirm', lambda _: False)
    langs = {l.id: l for l in api_copy.languoids()}
    with pytest.raises(AssertionError):
        _ = list(wikidata.Wikidata().iterupdated(langs.values()))



from clldutils.jsonlib import load

from pyglottolog.metadata import prepare_release


def test_read_editors(api_copy):
    prepare_release(api_copy, '3.3')
    zenodo = load(api_copy.path('.zenodo.json'))
    assert zenodo['creators'][1]['affiliation'] == 'University Uppsala'
    assert zenodo['description'] == '<p>, C &amp; Hammarstr&ouml;m, Harald &amp; Forkel, Robert. '\
                                    '1999. Glottolog 3.3. ' \
                                    'Jena: Max Planck Institute for the Science of Human History. '\
                                    '(Available online at ' \
                                    '<a href="https://glottolog.org">https://glottolog.org</a>)</p>'



import pytest


def test_legacy_import():
    from pyglottolog import api
    from pyglottolog import Glottolog
    assert api.Glottolog is Glottolog


def test_glottolog_invalid_repos(tmpdir):
    from pyglottolog import Glottolog
    with pytest.raises(ValueError, match=r'missing tree dir'):
        Glottolog(str(tmpdir))

    tmpdir.join('languoids').mkdir()
    tmpdir.join('languoids', 'tree').mkdir()

    with pytest.raises(ValueError, match=r'missing references subdir'):
        Glottolog(str(tmpdir))


def test_cache(caching_api, api):
    lang = api.languoid('abcd1234')
    assert lang is not api.languoid('abcd1234')

    lang = caching_api.languoid('abcd1234')
    assert lang is caching_api.languoid('abcd1234')
    assert caching_api.languoid('aaa') is lang
    l2 = caching_api.languoid('abc')
    assert l2.id == 'abcd1235'
    assert l2.ancestors


def test_cache_languoids(caching_api):
    assert 'aaa' not in caching_api.cache
    assert 'abcd1234' not in caching_api.cache
    _ = list(caching_api.languoids())
    assert 'aaa' in caching_api.cache
    assert 'abcd1234' in caching_api.cache


def test_refs_by_languoid(api):
    res = api.refs_by_languoid()
    assert len(res[0]) == 2
    assert len(res[1]) == 7
    res = api.refs_by_languoid('a')
    assert len(res[0]) == 1


def test_editors(api):
    eids = [
        e.id for e in sorted(api.editors.values(), key=lambda i: int(i.ord))
        if e.current]
    assert eids[0] == 'hammarstroem'


def test_paths(api):
    assert api.ftsindex


def test_languoid(api):
    assert api.languoid('abc').name == 'language'


def test_descendants_from_nodemap(api):
    nodemap = {n.id: n for n in api.languoids()}
    l = api.languoid('abcd1234')
    assert len(l.descendants_from_nodemap(nodemap)) == 3
    assert len(l.descendants_from_nodemap(nodemap, level=api.languoid_levels.language)) == 2
    assert len(l.descendants_from_nodemap(nodemap, level='dialect')) == 1

    with pytest.raises(ValueError):
        l.descendants_from_nodemap(nodemap, level='dialects')


def test_languoids(api):
    assert len(list(api.languoids())) == 8
    assert len(list(api.languoids(maxlevel=api.languoid_levels.family))) == 2
    assert len(list(api.languoids(maxlevel=0))) == 3
    assert len(list(api.languoids(maxlevel=0, exclude_pseudo_families=True))) == 2
    assert len(list(api.languoids(maxlevel=api.languoid_levels.language))) == 6
    assert len(api.languoids_by_code()) == 11
    assert api.languoids_by_code(nodes={}) == {}
    assert 'NOCODE_Family-name' in api.languoids_by_code()


def test_newick_tree(api):
    assert api.newick_tree(start='abcd1235') == \
        "('dialect [abcd1236]':1)'language [abcd1235][abc]-l-':1;"
    assert api.newick_tree(start='abcd1235') == \
        api.newick_tree(start='abcd1235', nodes={l.id: l for l in api.languoids()})
    assert api.newick_tree(start='abcd1235', template='{l.id}') == "(abcd1236:1)abcd1235:1;"
    assert set(api.newick_tree().split('\n')) == {
        "(('isolate {dialect} [dial1234]':1)'isolate [isol1234]-l-':1)'isolate [isol1234]':1;",
        "(('dialect [abcd1236]':1)'language [abcd1235][abc]-l-':1,"
        "'language2 [abcd1237]-l-':1)'family [abcd1234][aaa]':1;"
    }


def test_hhtypes(api):
    assert len(api.hhtypes) == 16


def test_load_triggers(api):
    assert len(api.triggers) == 2


def test_macroarea_map(api):
    assert api.macroarea_map['abc'] == 'Eurasia'



import pathlib

import pytest


from pyglottolog.languoids import (Languoid,
    Glottocodes, Glottocode, Country, Reference,
    ClassificationComment, EthnologueComment, Link)


def test_legacy_imports():
    from pyglottolog import objects
    assert objects.Glottocode is Glottocode
    assert objects.Reference is Reference


def test_Link():
    assert Link.from_(dict(url='xyz')).label is None
    assert Link.from_('[label](url)').url == 'url'
    assert 'url' in Link.from_('[label](url)').__json__()
    assert Link.from_string('abc').to_string() == 'abc'

    with pytest.raises(TypeError):
        Link.from_(5)


def test_Glottocodes(tmpdir):
    json = tmpdir / 'glottocodes.json'
    json.write_text('{}', encoding='ascii')

    glottocodes = Glottocodes(str(json))
    gc = glottocodes.new('a', dry_run=True)
    assert gc.startswith('aaaa')
    assert gc not in glottocodes
    gc = glottocodes.new('a')
    assert gc in glottocodes
    # make sure it's also written to file:
    assert gc in Glottocodes(str(json))
    assert len(list(Glottocodes(str(json)))) == 1


@pytest.mark.parametrize('input_, valid', [
    ('abcd1234', True),
    ('a12d3456', True),
    ('abcd123', False),
    ('12d3456', False),
    ('aNOCODE', False),
    ('NOCODE_abd', False),
    ('nocode', False),
])
def test_pattern(input_, valid, _match=Glottocode.pattern.match):
    assert (_match(input_) is not None) == valid


@pytest.mark.parametrize('text, expected_id, expected_str', [
    ('Germany', 'DE', 'DE'),
    ('Russian Federation (RU)', 'RU', 'RU'),
    ('RU', 'RU', 'RU'),
])
def test_Country_from_text(text, expected_id, expected_str):
    country = Country.from_text(text)
    assert country.id == expected_id
    assert str(country) == expected_str


def test_Country_from_name():
    assert Country.from_name('abcdefg') is None


def test_Country_from_id():
    assert Country.from_id('abcdefg') is None


def test_Glottocode_validation():
    with pytest.raises(ValueError):
        Glottocode('a2')


def test_Glottocode_ordering():
    assert sorted([Glottocode('abcd1235'), Glottocode('abcd1234')])[0] == Glottocode('abcd1234')
    assert Glottocode('zzzz9999') > Glottocode('abcd1234')
    assert Glottocode('abcd1234') <= Glottocode('abcd1234')


def test_Reference():
    ref = Reference('bib:key', '12-34', 'German')
    assert '{0}'.format(ref) == '**bib:key**:12-34<trigger "German">'
    Reference.from_list(['{0}'.format(ref)])

    with pytest.raises(ValueError):
        Reference.from_list(['abc'])

    match = Reference.pattern.match('**bib:k(e)y(**)')
    assert match.group('key') == 'bib:k(e)y'
    assert match.group('endtag') == '(**)'


def test_ClassificationComment(mocker):
    cc = ClassificationComment(family='**bib:key**')
    log = mocker.Mock()
    cc.check(mocker.Mock(), [], log)
    assert log.error.called
    log = mocker.Mock()
    cc.check(mocker.Mock(), ['bib:key'], log)
    assert not log.error.called
    cc = ClassificationComment(sub='A comment about **bib:key**:10-12', subrefs=['**bib:key**:20'])
    refs = cc.merged_refs('sub')
    assert len(refs) == 1
    assert refs[0].pages == '10-12;20'


def test_EthnologueComment(mocker):
    with pytest.raises(ValueError):
        EthnologueComment('abc', 't')

    with pytest.raises(ValueError):
        EthnologueComment('abc', 'missing', 'E15')

    with pytest.raises(ValueError):
        EthnologueComment('abc', 'missing', 'E16')

    with pytest.raises(ValueError):
        EthnologueComment('abc', 'missing', 'E16', '\u00e4\u00f6\u00fc'.encode('utf-8'))

    log = mocker.Mock()
    ec = EthnologueComment('abc', 'missing', 'E16', 'abc')
    ec.check(mocker.Mock(), [], log)
    assert not log.error.called

    log = mocker.Mock()
    ec = EthnologueComment('abc', 'missing', 'E16', '**bib:key**')
    ec.check(mocker.Mock(), [], log)
    assert log.error.called


def test_ancestors(api):
    lang = api.languoid('dial1234')
    assert 'isol1234' in ','.join(repr(a) for a in lang.ancestors)


def test_endangerment(api):
    lang = api.languoid('abcd1235')
    assert lang.endangerment.status == api.aes_status.sleeping
    assert 'source' in lang.endangerment.__json__()


def test_timespan(api, tmpdir, recwarn):
    assert api.languoid('abcd1235').timespan == None

    with pytest.raises(ValueError):
        api.languoid('abcd1235').timespan = (1,)

    l = api.languoid('abcd1235')
    l.timespan = (-1, 300)
    assert l.timespan == (-1, 300)

    p = pathlib.Path(str(tmpdir))
    l.write_info(p)
    assert '-0001-01-01/0300-01-01' in p.joinpath('abcd1235', 'md.ini').read_text(encoding='utf8')

    l.timespan = (-10000, 2000)
    assert recwarn.pop(UserWarning)


def test_Level(api):
    assert api.languoid_levels.dialect > api.languoid_levels.language
    assert api.languoid_levels.language == api.languoid('abcd1235').level
    with pytest.raises(ValueError):
        api.languoid_levels.get('abcde')


def test_Category(api):
    assert api.languoid('book1243').category == api.language_types.bookkeeping.category
    assert api.languoid('abcd1235').category == api.language_types.spoken_l1.category
    assert api.languoid('dial1234').category == 'Dialect'


def test_sources(api):
    assert api.languoid('book1242').sources == []
    with pytest.raises(AssertionError):
        api.languoid('book1242').sources = [1]
    api.languoid('book1242').sources = [Reference('key')]


def test_ethnologue_comment(api):
    assert api.languoid('book1243').ethnologue_comment.comment_type == 'missing'
    assert api.languoid('book1243').ethnologue_comment.__json__()


def test_classification_comment(api):
    assert api.languoid('abcd1234').classification_comment


def test_coordinates_setter(api, tmpdir):
    l = api.languoid('abcd1234')
    l.latitude = 1.12345678
    l.longitude = 1.123456789
    l.write_info(pathlib.Path(str(tmpdir)))
    ini = pathlib.Path(str(tmpdir)).joinpath('abcd1234', 'md.ini').read_text(encoding='utf8')
    assert 'latitude = 1.12346' in ini and 'longitude = 1.12346'


def test_classification_setter(api, tmpdir):
    l = api.languoid('abcd1234')
    l.cfg.set('classification', 'familyrefs', ['abc', 'def'])
    l.write_info(pathlib.Path(str(tmpdir)))
    ini = pathlib.Path(str(tmpdir)).joinpath('abcd1234', 'md.ini').read_text(encoding='utf8')
    assert '\tabc\n\tdef' in ini


def test_Languoid_sorting(api):
    assert api.languoid('abcd1235') < api.languoid('abcd1236')
    assert api.languoid('abcd1236') >= api.languoid('abcd1235')


def test_factory_without_api(api_copy):
    f = Languoid.from_dir(api_copy.tree / 'abcd1234', _api=api_copy)
    l = Languoid.from_dir(api_copy.tree / f.id / 'abcd1235')
    assert len(l.macroareas) == 0  # No API passed at initialization => no macroareas!


def test_factory(tmpdir, api_copy):
    f = Languoid.from_dir(api_copy.tree / 'abcd1234', _api=api_copy)
    assert f.category == 'Family'
    l = Languoid.from_dir(api_copy.tree / f.id / 'abcd1235', _api=api_copy)
    assert l.name == 'language'
    assert 'abcd1235' in repr(l)
    assert 'language' in '%s' % l
    assert l.level == api_copy.languoid_levels.language
    assert l.latitude == pytest.approx(0.5)
    assert l.longitude == pytest.approx(-30)
    l.latitude, l.longitude = 1.0, 1.0
    assert l.latitude == pytest.approx(1.0)
    assert l.longitude == pytest.approx(1.0)
    assert l.iso_code == 'abc'
    l.iso_code = 'cde'
    assert l.iso == 'cde'
    assert l.hid == 'abc'
    l.hid = 'abo'
    assert l.hid == 'abo'
    assert l.id == 'abcd1235'

    assert len(l.macroareas) == 2
    l.macroareas = [api_copy.macroareas.pacific]
    assert l.macroareas == [api_copy.macroareas.get('Papunesia')]

    l.countries = api_copy.countries[:2]
    assert len(l.countries) == 2

    assert l.parent == f
    assert l in f.children
    assert l.children[0].family == f
    l.write_info(str(tmpdir))
    assert (tmpdir / 'abcd1235').exists()
    assert isinstance(api_copy.languoid('abcd1235').iso_retirement.asdict(), dict)
    assert l.classification_comment is None
    assert l.names == {}
    l.cfg['altnames'] = {'glottolog': 'xyz'}
    assert 'glottolog' in l.names
    assert l.identifier == {}
    l.cfg['identifier'] = {'multitree': 'xyz'}
    assert 'multitree' in l.identifier


def test_isolate(api):
    l = Languoid.from_dir(api.tree / 'isol1234')
    assert l.isolate
    assert l.parent is None
    assert l.family is None


def test_closest_iso(api):
    assert api.languoid('abcd1235').closest_iso() == 'abc'
    assert api.languoid('abcd1236').closest_iso() == 'abc'


def test_attrs(api):
    l = Languoid.from_name_id_level(api.tree, 'name', 'abcd1235', 'language', hid='NOCODE')
    l.name = 'other'
    assert l.name == 'other'
    with pytest.raises(AttributeError):
        l.glottocode = 'x'
    with pytest.raises(AttributeError):
        l.id = 'x'
    assert l.id == l.glottocode
    assert l.hid == 'NOCODE'


def test_iter_descendants(api):
    children = [l.id for l in api.languoid('abcd1234').iter_descendants()]
    assert set(children) == {'abcd1235', 'abcd1236', 'abcd1237'}
    


import shlex
import shutil
import logging

import pytest

from pyglottolog.__main__ import admin_main


@pytest.fixture
def _main(api_copy, mocker):
    def f(*args, **kw):
        kw.setdefault('log', mocker.Mock())
        if len(args) == 1 and isinstance(args[0], str):
            args = shlex.split(args[0])
        admin_main(args=['--repos', str(api_copy.repos)] + list(args), **kw)
    return f


def test_release(_main):
    with pytest.raises(AssertionError):
        _main('release --version 1.0.1.dev0')
    _main('release --version 2.7')


def test_update_links(_main, capsys):
    _main('updatelinks none')
    out, _ = capsys.readouterr()
    assert '0 languoids updated' in out


def test_lff(capsys, _main, api_copy, encoding='utf-8'):
    _main('tree2lff')

    dff = api_copy.build_path('dff.txt')
    dfftxt = dff.read_text(encoding=encoding).replace('dialect', 'Dialect Name')
    dff.write_text(dfftxt, encoding=encoding)
    _main('lff2tree')
    assert 'git status' in capsys.readouterr()[0]
    assert api_copy.languoid('abcd1236').name == 'Dialect Name'
    # Old language and dialect names are retained as alternative names:
    assert 'dialect' in api_copy.languoid('abcd1236').names['glottolog']

    _main('tree2lff')
    _ = dff.read_text(encoding=encoding)


def test_index(api_copy, _main):
    _main('langindex')
    assert len(list(api_copy.repos.joinpath('languoids').glob('*.md'))) == 10


def test_update_macroareas(_main, capsys):
    _main('updatemacroareas')
    out, _ = capsys.readouterr()
    assert '1 dialects updated' in out


def test_update_sources(_main, capsys):
    _main('updatesources')
    out, _ = capsys.readouterr()
    assert '2 languoids updated' in out


def test_check(capsys, _main, mocker, api_copy):
    _main('check --bib-only')

    log = mocker.Mock()
    _main('check --tree-only', log=log)
    assert 'family' in capsys.readouterr()[0]
    msgs = [a[0] for a, _ in log.error.call_args_list]
    assert any('unregistered glottocode' in m for m in msgs)
    assert any('missing reference' in m for m in msgs)
    assert len(msgs) == 28

    shutil.copytree(
        api_copy.tree / 'abcd1234' / 'abcd1235',
        api_copy.tree / 'abcd1235')

    log = mocker.Mock()
    _main('check --tree-only', log=log)
    msgs = [a[0] for a, _ in log.error.call_args_list]
    assert any('duplicate glottocode' in m for m in msgs)
    assert len(msgs) == 30

    (api_copy.tree / 'abcd1235').rename(api_copy.tree / 'abcd1238')
    log = mocker.Mock()
    _main('check --tree-only', log=log)
    msgs = [a[0] for a, _ in log.error.call_args_list]
    assert any('duplicate hid' in m for m in msgs)
    assert len(msgs) >= 9


def test_check_2(_main, mocker, caplog):
    mocker.patch('pyglottolog.iso.check_lang', lambda _, i, l, **kw: ('warn', l, 'xyz'))
    _main('check --tree-only', log=logging.getLogger(__name__))
    for record in caplog.records:
        print(record.message)


def test_monster(capsys, _main):
    _main('bib')
    assert capsys.readouterr()[0]



"""
Integrate data from ELCat's CLDF dataset into Glottolog.
"""
import re
import datetime
import itertools
import collections
import typing

from pycldf import Dataset
from pycldf.orm import Language

from pyglottolog.languoids.models import Country, Glottocode
from .util import LinkProvider

CFG_ID_NAME = "endangeredlanguages"
DOMAIN = "{}.com".format(CFG_ID_NAME)
LINK_TYPE = "elcat"
CLDF_DATASET = "https://raw.githubusercontent.com/" \
               "cldf-datasets/elcat/main/cldf/StructureDataset-metadata.json"


def format_comment(ellang: Language, lei: typing.Union[dict, str]) -> str:
    comment = "{} ({}{}) = {}".format(
        ellang.cldf.name,
        ellang.id,
        '-{}'.format(ellang.cldf.iso639P3code) if ellang.cldf.iso639P3code else '',
        (lei['Value'] if isinstance(lei, dict) else lei).replace('()', '').strip(),
    )
    if isinstance(lei, dict):
        if lei['Comment']:
            comment += ' ({})'.format(lei['Comment'])
        if lei['Source']:
            comment += ' [{}]({})'.format(
                lei['Source'][0].refkey(year_brackets=None), 'elcat:' + lei['Source'][0].id)
    return comment


def format_date() -> str:
    return datetime.datetime.now().isoformat().split('.')[0]


class ElCat(LinkProvider):
    __inactive__ = True
    status_map = {}

    def __init__(self, repos=None):
        LinkProvider.__init__(self, repos=repos)
        self.languages = collections.defaultdict(list)
        cldf = Dataset.from_metadata(CLDF_DATASET)
        self.lei = {}
        for lg in cldf.objects('LanguageTable', Language):
            if lg.cldf.glottocode:
                self.languages[lg.cldf.glottocode].append(lg)
            elif 'Glottolog' in lg.data['code_authorities']:
                for code in lg.data['codes']:
                    if Glottocode.pattern.match(code):
                        self.languages[code].append(lg)
            elif 'ISO 639-3' in lg.data['code_authorities']:
                for code in lg.data['codes']:
                    if re.fullmatch('[a-z]{3}', code):
                        self.languages[code].append(lg)
            self.lei[lg.id] = lg.data['endangerment']
        self.status_map = {}
        for status in self.repos.aes_status.values():
            for name in status.elcat.split('/'):
                self.status_map[name] = status
        for (lid, pid), rows in itertools.groupby(
            sorted(
                cldf.iter_rows('ValueTable'), key=lambda r: (r['Language_ID'], r['Parameter_ID'])),
            lambda r: (r['Language_ID'], r['Parameter_ID'])
        ):
            if pid == 'LEI':
                rows = list(rows)
                stati = []
                for row in rows:
                    stati.append(self.map_status(row['Value']).name)
                    if (((self.lei[lid] is None)
                         or (stati[-1] == self.map_status(self.lei[lid]).name))  # noqa W503
                            and (row['preferred'] is True)):  # noqa W504
                        row['Source'] = [cldf.sources[src] for src in row['Source']]
                        self.lei[lid] = row
                        break
                else:
                    # We pick the first source reporting the same status as the one chosen by ELCat
                    # overall as reference:
                    for i, status in enumerate(stati):
                        if status == self.lei[lid]:
                            rows[i]['Source'] = [cldf.sources[src] for src in rows[i]['Source']]
                            self.lei[lid] = rows[i]

    def map_status(self, s):
        if s:
            return self.status_map[s.split('(')[0].strip().lower()]

    def iterupdated(self, languoids):
        for lang in languoids:
            changed, ellangs = False, []
            if lang.id in self.languages:
                ellangs = self.languages[lang.id]
            elif lang.iso and (lang.iso in self.languages):
                ellangs = self.languages[lang.iso]

            if ellangs:  # There are ELCat languages mapped to the Glottolog languoid.
                changed = lang.update_links(DOMAIN, [(l_.valueUrl(), l_.name) for l_ in ellangs])

                countries = {c for c in itertools.chain(*[l_.data['Countries'] for l_ in ellangs])}
                existing = {c.id for c in lang.countries}
                if not countries.issubset(existing):
                    lang.countries = [Country.from_id(c) for c in sorted(countries.union(existing))]
                    changed = True

                if len(ellangs) != 1:
                    # Assign best case status as multiple ELCat varieties mapped to one languoid
                    # means ELCat describes dialects. Then the entire language is as endangered as
                    # the least endangered dialect.
                    try:
                        ellang = sorted(
                            [lg for lg in ellangs if lg.data['endangerment']],
                            key=lambda e: self.map_status(e.data['endangerment']))[0]
                    except IndexError:
                        ellang = ellangs[0]
                else:
                    ellang = ellangs[0]
                    # Only add alternative names, if only one ElCat language matches!
                    changed = lang.update_names(
                        [ellang.cldf.name] + ellang.data['alt_names'], type_=LINK_TYPE) or changed

                    # Add missing coordinates
                    if (not lang.latitude) and ellang.cldf.latitude and ellang.cldf.longitude:
                        # Only add missing coordinates, if ElCat lists only one coordinate pair!
                        lang.latitude = ellang.cldf.latitude
                        lang.longitude = ellang.cldf.longitude
                        changed = True

                lei = self.lei[ellang.id]
                value = None
                if isinstance(lei, dict):
                    value = lei
                    lei = lei['Value']

                if lang.endangerment:
                    if lang.endangerment.source.id == 'ElCat':
                        if lei is None:
                            del lang.cfg['endangerment']
                        else:
                            lang.cfg['endangerment']['status'] = self.map_status(lei).name
                            lang.cfg['endangerment']['date'] = format_date()
                            lang.cfg['endangerment']['comment'] = format_comment(
                                ellang, value or lei)
                        changed = True
                elif self.map_status(lei):
                    lang.cfg['endangerment'] = dict(
                        status=self.map_status(lei).name,
                        date=format_date(),
                        comment=format_comment(ellang, value or lei),
                        source='ElCat')
                    changed = True

            else:
                changed = any([lang.update_links(DOMAIN, []),
                               lang.update_names([], type_=LINK_TYPE)])

            if changed:
                yield lang



"""
Write release info to .zenodo.json, CITATION.md and CONTRIBUTORS.md
"""
import shlex
import subprocess

import packaging.version

from pyglottolog.metadata import prepare_release


def register(parser):
    parser.add_argument('--version', help="version number without leading 'v'", default=None)


def run(args):
    try:
        branch = ''
        out = subprocess.check_output(shlex.split("git -C {} branch".format(args.repos.repos)))
        for line in out.decode('utf8').split('\n'):
            if line.startswith('*'):
                branch = line[1:].strip()
        assert branch.startswith('release')
    except subprocess.CalledProcessError:
        pass
    version = getattr(args.repos.publication.zenodo, 'version', args.version)
    assert not packaging.version.parse(version).is_prerelease, 'invalid release version number'
    print(prepare_release(args.repos, version=version))
